/*
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 * Or, point your browser to http://www.gnu.org/copyleft/gpl.html
 */

#include <stdint.h>
#include "stioctl.h"

#define	TS_SIZE	188
#define	STPTI_TIMEOUT	1000 // 10*1000

typedef struct {
  unsigned int sthandle;
  unsigned int vidhandle;
  unsigned int audhandle;
  unsigned int pcrhandle;
  unsigned int pmthandle;
  unsigned int vpid;
  unsigned int apid;
  unsigned int pcrpid;
} st_slothandles_t;


int open_stapi(void);
int allocate_buffer(int, int, void**, int);
int free_buffer(int, void*);
int set_buffer(int, int, void*);
int unset_buffer(int, int);
void get_slot_handles(st_slothandles_t*);

void reset_pat_counter(void);
void create_pat_pmt(unsigned char*, uint16_t, uint16_t, uint16_t);

